/*
 *  Copyright (C) 2006-2009 Leandro Nini
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "audio.h"
#include "outputFactory.h"
#include "settings.h"

#define DECIMALS 10
//#define PROFILE

/*****************************************************************/

FXint audioThread::run() {
	_audio->loop();
	return 1;
}

/*****************************************************************/

#ifdef PROFILE
namespace FX {
extern FXAPI FXlong fxgetticks();
}
#endif

void audio::loop() {

	while (_playing) {
#ifdef PROFILE
	FXlong ticks=FX::fxgetticks();
#endif
		int size=_input->fillBuffer(
			_converter?_converter->buffer():_output->buffer(),
			_converter?_converter->bufSize():_bufferSize,
			_seconds);

		if (!size) {
			_guiMsg->setData((void*)SONG_ENDED);
			_guiMsg->signal();
			if (_preload) {
				_input=_preload;
				_preload=0;
				_seconds=0;
				continue;
			} else
				break;
		}

		if (_converter)
			size=_converter->convert(_output->buffer(), size);

#if (FOX_BIGENDIAN == 1)
		//Swap bytes on big endian machines
		if (_input->precision()==2) {
			FXshort *buf=(FXshort*)_output->buffer();
			FXshort *end=buf+(size/2);
			do {
				FXShort tmp=*buf;
				*buf++=((tmp&0x00FF)<<8)&((tmp&0xFF00)>>8);
			} while (buf<end);
		}
#endif
#ifdef HAVE_BS2B
		if (_bs2bdp) {
#if BS2B_VERSION_MAJOR == 3
			switch (_input->precision()) {
			case 1:
			{
				uint8_t *buf=(uint8_t*)_output->buffer();
				bs2b_cross_feed_u8(_bs2bdp, buf, size/2);
			}
			break;
			case 2:
			{
				int16_t *buf=(int16_t*)_output->buffer();
				bs2b_cross_feed_s16le(_bs2bdp, buf, size/4);
			}
			break;
			}
#else
			short *buf=(short*)_output->buffer();
			const short *end=buf+(size/2);
			do {
				bs2b_cross_feed_16(_bs2bdp, buf);
				buf+=2;
			} while (buf<end);
#endif
		}
#endif
#ifdef PROFILE
	ticks=FX::fxgetticks()-ticks;
	fxmessage("Ticks %d\n", (int)ticks);
#endif

		if (!_output->write(_output->buffer(), size)) {
			fxmessage("Output error\n");
			_playing=false;
			break;
		}

		_buffers-=(1<<DECIMALS);
		if (_buffers<0) {
			_buffers+=_bufPerSec;
			_seconds++;
			if (_seconds!=_input->time()-5)
				_guiMsg->setData((void*)UPD_TIME);
			else
				_guiMsg->setData((void*)PRELOAD_SONG);
			_guiMsg->signal();
		}
	}
}

/*****************************************************************/

audio::audio(FXApp *app, FXObject *o, FXSelector guiMsg) :
	_app(app),
	_card(0),
	_input(0),
	_preload(0),
	_state(STOP),
	_playing(false),
	_seconds(0)
#ifdef HAVE_BS2B
	,_bs2bdp(0)
#endif
	{

	_guiMsg=new FXGUISignal(app, o, guiMsg);

	const FXString apiName=_app->reg().readStringEntry("Audio Settings", "api");
	const FXString card=_app->reg().readStringEntry("Audio Settings", "card");
	_volume=_app->reg().readIntEntry("Audio Settings", "volume", 50);

	_output=OFACTORY->get(_app, apiName.text());

	for (int i=0; i<_output->devices(); i++) {
		if (!compare(card, _output->device(i))) {
			_card=i;
			break;
		}
	}
}

audio::~audio() {

	stop();

	_app->reg().writeStringEntry("Audio Settings", "api", _output->name());
	_app->reg().writeStringEntry("Audio Settings", "card", _output->device(_card).text());
	_app->reg().writeIntEntry("Audio Settings", "volume", _volume);

	delete _output;
	delete _guiMsg;
}

bool audio::play(input* i) {

	if ((!_output) || (!i->songLoaded()) || (_state==PLAY))
		return false;

	FXTRACE((10, "audio::play\n"));

	_input=i;

	int sampleRate=_input->samplerate();
	FXTRACE((10, "Setting parameters %d:%d:%d\n", sampleRate, _input->channels(), _input->precision()<<3));

	_bufferSize=_output->open(_card, sampleRate, _input->channels(), _input->precision());
	if (!_bufferSize)
		return false;

	FXTRACE((10,"Output samplerate %d\n", sampleRate));

	// Check if soundcard supports requested samplerate
	_converter=(sampleRate!=_input->samplerate())?
		new converter(_input->samplerate(), sampleRate, _bufferSize, _input->channels(), _input->precision()):0;

	_bufPerSec=((sampleRate<<DECIMALS)*_input->channels()*_input->precision())/_bufferSize;
	_buffers=_bufPerSec;
#ifdef HAVE_BS2B
	if (SETTINGS->bs2b() && _input->channels()==2
#if BS2B_VERSION_MAJOR == 2
		&& _input->precision()==2
#endif
	) {
		_bs2bdp=bs2b_open();
		if (_bs2bdp) {
			FXTRACE((10, "bs2b enabled\n"));
			bs2b_set_srate(_bs2bdp, sampleRate);
			bs2b_set_level(_bs2bdp, BS2B_DEFAULT_CLEVEL);
		}
	}
#endif
	_output->volume(_volume);

	_thread=new audioThread(this);

	_playing=true;
	_state=PLAY;
	_thread->start();

	return true;
}

void audio::pause() {

	switch (_state) {
	case PLAY:
		FXTRACE((10, "Pause\n"));
		_output->pause();
		_playing=false;
		_thread->join();
		_state=PAUSE;
		break;
	case PAUSE:
		FXTRACE((10, "Unpause\n"));
		_playing=true;
		_thread->start();
		_state=PLAY;
		break;
	}
}

bool audio::stop() {

	if ((!_output) || (_state==STOP))
		return false;

	FXTRACE((10, "audio::stop\n"));

	_playing=false;
	_thread->join();

	_output->close();

	delete _thread;

	_seconds=0;
	_state=STOP;
#ifdef HAVE_BS2B
	if (_bs2bdp) {
		bs2b_close(_bs2bdp);
		_bs2bdp=0;
	}
#endif
	if (_converter)
		delete _converter;

	_input->rewind();

	return true;
}

void audio::volume(const int vol) {

	_volume=vol;
	if (_output)
		_output->volume(_volume);
}

bool audio::gapless(input* const i) {

	if ((_input->samplerate()==i->samplerate()) &&
	(_input->channels()==i->channels()) &&
	(_input->precision()==i->precision())) {
		_preload=i;
		return true;
	} else {
		_preload=0;
		return false;
	}
}

void audio::config(FXWindow* win) {

	audioConfig cWin(win);
	cWin.api(_output->name());
	cWin.card(_card);

	if (cWin.execute(PLACEMENT_OWNER)) {
		const bool playing=_playing;
		stop();
		delete _output;
		_card=cWin.card();
		_output=OFACTORY->get(_app, cWin.api());
		if (playing)
			play(_input);
	}
}

/*****************************************************************/

FXDEFMAP(audioConfig) audioConfigMap[] = {
	// Message_Type		ID				Message_Handler
	FXMAPFUNC(SEL_COMMAND,	audioConfig::ID_API,		audioConfig::onCmdApi),
};

FXIMPLEMENT(audioConfig, configWindow, audioConfigMap, ARRAYNUMBER(audioConfigMap))

audioConfig::audioConfig(FXWindow* win) :
	configWindow(win, "Audio") {

	new FXLabel(matrix(), tr("API"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
	_apiList=new FXListBox(matrix(), this, audioConfig::ID_API, LAYOUT_FILL|FRAME_NORMAL);

	const int apis=OFACTORY->num();
	for (int i=0; i<apis; i++)
		_apiList->appendItem(OFACTORY->name(i));

	_apiList->setNumVisible(apis);

	new FXLabel(matrix(), tr("Card"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
	_cardList=new FXListBox(matrix(), 0, 0, LAYOUT_FILL|FRAME_NORMAL);
}

long audioConfig::onCmdApi(FXObject*, FXSelector, void* ptr) {

	const FXint val=(FXint)(FXival)ptr;
	setCards(OFACTORY->get(getApp(), val));

	return 1;
}

void audioConfig::api(const char* apiName) {

	FXTRACE((10, "Api: %s\n", apiName));

	_apiList->setCurrentItem(_apiList->findItem(apiName));

	setCards(OFACTORY->get(getApp(), apiName));
}

void audioConfig::setCards(const output* const audio) {

	if (!audio)
		return;

	_cardList->clearItems();

	const int devices=audio->devices();
	FXTRACE((10, "Devices: %d\n", devices));
	for (int i=0; i<devices; i++) {
		_cardList->appendItem(audio->device(i));
	}

	delete audio;

	_cardList->setNumVisible((devices>5)?5:devices);
}

void audioConfig::card(const int card) {

	if (card<_cardList->getNumItems())
		_cardList->setCurrentItem(card);
}
