/*
 *  Copyright (C) 2006-2009 Leandro Nini
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "game.h"

#ifdef BUILD_GME

#include <FX88591Codec.h>
#include "utils.h"

#ifdef HAVE_LIBZ
#  define EXT	"*.(ay|gbs|gym|hes|kss|nsfe|nsf|sap|spc|vgm|vgz)"
#else
#  define EXT	"*.(ay|gbs|gym|hes|kss|nsfe|nsf|sap|spc|vgm)"
#endif

// ASMA path to STIL.
#define ASMA_STIL	"/Docs/STIL.txt"

// ASMA path to BUGlist.
#define ASMA_BUGLIST	"/Docs/Bugs.txt"

#define CREDITS		"Game_Music_Emu library 0.5.2\nCopyright \302\251 Shay Green."
#define LINK		"http://www.slack.net/~ant/"

const char game::name[]="Gme";

/*****************************************************************/

int game::fillBuffer(char* buffer, const unsigned int bufferSize, const unsigned int seconds) {

	if (_emu->track_ended())
		return 0;

	_emu->play((bufferSize>>1), (Music_Emu::sample_t*)buffer);
	return bufferSize;
}

/*****************************************************************/

bool game::supports(const char* fileName) {

	return FXPath::match(FXString(EXT), FXPath::name(fileName), FILEMATCH_FILE_NAME|FILEMATCH_NOESCAPE|FILEMATCH_CASEFOLD);
}

inline FXString game::ext() const { return FXString(EXT); }

game::game(FXApp* a) :
	inputPlugin(a, name),
	_emu(0),
	_stil(0) {

	_settings.samplerate=load("Samplerate", 44100);
	_settings.asmaPath=load("ASMA", FXString::null);

	openAsma(_settings.asmaPath);
}

game::~game() {

	close();

	save("Samplerate", _settings.samplerate);
	save("ASMA", _settings.asmaPath);

	delete _stil;
}

bool game::open(const FXString& fileName) {

	close();

	gme_type_t fileType;
	if (!check(gme_identify_file(fileName.text(), &fileType)))
		return false;

	FXTRACE((10, "System %s\n", fileType->system));

	_emu=fileType->new_emu();
	if (!check(_emu->set_sample_rate(_settings.samplerate)))
		return false;
	if (!check(_emu->load_file(fileName.text())))
		return false;
	if (!check(_emu->start_track(0)))
		return false;

	_emu->load_m3u(FXStringFormat("%s.m3u",FXPath::stripExtension(fileName).text()).text());

	getInfo();

	_hasStilInfo=_stil && !comparecase(FXPath::extension(fileName), "sap");
	if (_hasStilInfo) {
		FXTRACE((10, "Retrieving STIL info\n"));
		FXString comment=FXString(_stil->getAbsGlobalComment(fileName.text()));
		if (!comment.empty())
			comment.append('\n');
		comment.append(FXString(_stil->getAbsEntry(fileName.text())));
		FXString bug=FXString(_stil->getAbsBug(fileName.text()));
		if (!bug.empty()) {
			comment.append('\n');
			comment.append(bug);
		}
		if (!comment.empty()) {
			const FX88591Codec codec;
			addInfo(COMMENT, codec.mb2utf(comment.text()).text());
		}
	}

	songLoaded(true);
	return true;
}

void game::getInfo() {

	track_info_t ti;
	const char* err=_emu->track_info(&ti);
	if (err) {
		fxmessage("Warning: %s\n", err);
		return;
	}

	//TODO maybe clear all info
	delInfo(TITLE);
	addInfo(TITLE, ti.song);
	addInfo(ARTIST, ti.author);
	addInfo("Copyright", ti.copyright);
	addInfo("System", ti.system);
	addInfo("Game", ti.game);
	addInfo("Dumper", ti.dumper);
	if (!_hasStilInfo)
		addInfo(COMMENT, ti.comment);

	time(ti.length/1000);
}

bool game::check(const char* error) {

	if (error) {
		delPtr(_emu);
		fxmessage("Error: %s\n", error);
		return false;
	}
	return true;
}

void game::close() {

	delPtr(_emu);

	songLoaded(false);
}

const bool game::rewind() {

	if (_emu) {
		_emu->start_track(_emu->current_track());
		return true;
	}
	return false;
}

const bool game::config(FXWindow* win) {

	const FXString oldAsmaPath=_settings.asmaPath;
	gameConfig cWin(win, _settings);
	if (cWin.execute(PLACEMENT_OWNER)) {
		_settings=cWin.settings();
		if (compare(_settings.asmaPath, oldAsmaPath))
			openAsma(_settings.asmaPath);
		return true;
	}
	return false;
}

bool game::subtune(const FXuint i) {

	if (_emu && (i>0) && (i<=_emu->track_count())) {
		_emu->start_track(i-1);
		getInfo();
		return true;
	}
	return false;
}

void game::openAsma(const FXString& asmaPath) {

	if (asmaPath.empty())
		return;

	if (!_stil)
		_stil=new STIL(ASMA_STIL, ASMA_BUGLIST);

	if (!_stil->setBaseDir(asmaPath.text())) {
		fxmessage("%s\n", _stil->getErrorStr());
		delPtr(_stil);
	}
}

/*****************************************************************/

FXDEFMAP(gameConfig) gameConfigMap[] = {
	// Message_Type		ID				Message_Handler
	FXMAPFUNC(SEL_COMMAND,	gameConfig::ID_SAMPLERATE,	gameConfig::onCmdSamplerate),
	FXMAPFUNC(SEL_COMMAND,	gameConfig::ID_ASMA,		gameConfig::onCmdAsma),
};

FXIMPLEMENT(gameConfig, configWindow, gameConfigMap, ARRAYNUMBER(gameConfigMap))

gameConfig::gameConfig(FXWindow* win, gmeConfig_t settings) :
	configWindow(win, game::name, CREDITS, LINK),
	_settings(settings) {

	FXint val;

	_asmaDir.connect(_settings.asmaPath);

	new FXLabel(matrix(), tr("Samplerate"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
	FXListBox *freqBox=new FXListBox(matrix(), this, gameConfig::ID_SAMPLERATE, LAYOUT_FILL|FRAME_NORMAL);
	freqBox->fillItems("11025\n22050\n44100\n48000");
	freqBox->setNumVisible(4);
	switch (_settings.samplerate) {
	case 11025:
		val=0;
		break;
	case 22050:
		val=1;
		break;
	default:
	case 44100:
		val=2;
		break;
	case 48000:
		val=3;
		break;
	}
	freqBox->setCurrentItem(val);

	FXHorizontalFrame *hf=new FXHorizontalFrame(extraBottom());
	new FXLabel(hf, tr("ASMA path:"));
	new FXTextField(hf, 40, &_asmaDir, FXDataTarget::ID_VALUE, TEXTFIELD_READONLY|TEXTFIELD_NORMAL);
	new FXButton(hf, tr("&Browse\tSelect ASMA directory"), 0, this, gameConfig::ID_ASMA, FRAME_RAISED|FRAME_THICK|JUSTIFY_LEFT|ICON_BEFORE_TEXT);
}

long gameConfig::onCmdSamplerate(FXObject*, FXSelector, void* ptr) {

	const FXint val=(FXint)(FXival)ptr;
	switch (val) {
	case 0:
		_settings.samplerate=11025;
		break;
	case 1:
		_settings.samplerate=22050;
		break;
	case 2:
		_settings.samplerate=44100;
		break;
	case 3:
		_settings.samplerate=48000;
		break;
	}
	return 1;
}

long gameConfig::onCmdAsma(FXObject*, FXSelector, void*) {

	FXDirDialog dirDialog(this, tr("Select ASMA directory"));
	if (dirDialog.execute())
		_settings.asmaPath=dirDialog.getDirectory();

	return 1;
}

#endif
