/*
 *  Copyright (C) 2007-2009 Leandro Nini
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "hvl.h"

#ifdef BUILD_HVL

#include "utils.h"

#define EXT	"*.(ahx|hvl),(ahx|hvl).*"

#define CREDITS		"HivelyTracker\nDexter/Abyss, Xeron/IRIS, and Pieknyman"
#define LINK		"http://www.hivelytracker.com/"

const char hvl::name[]="Hively";

/*****************************************************************/

int hvl::fillBuffer(char* buffer, const unsigned int bufferSize, const unsigned int seconds) {

	unsigned int offset=0;

	if (_left) {
		memcpy(buffer, _buffer+_size-_left, _left);
		offset=_left;
		_left=0;
	}

	do {
		hvl_DecodeFrame(_tune, ((int8*)buffer)+offset, ((int8*)buffer)+offset+2, 4);
		offset+=_size;
	} while (offset<bufferSize-_size);

	const int left=bufferSize-offset;
	if (left) {
		hvl_DecodeFrame(_tune, (int8*)_buffer, ((int8*)_buffer)+2, 4);
		memcpy(buffer+offset, _buffer, left);
		_left=_size-left;
	}

	return bufferSize;
}

/*****************************************************************/

bool hvl::supports(const char* fileName) {

	return FXPath::match(FXString(EXT), FXPath::name(fileName), FILEMATCH_FILE_NAME|FILEMATCH_NOESCAPE|FILEMATCH_CASEFOLD);
}

inline FXString hvl::ext() const { return FXString(EXT); }

hvl::hvl(FXApp* a) :
	inputPlugin(a, name),
	_tune(0),
	_buffer(0) {

	_settings.samplerate=load("Samplerate", 44100);

	hvl_InitReplayer();
}

hvl::~hvl() {

	close();

	save("Samplerate", _settings.samplerate);
}

bool hvl::open(const FXString& fileName) {

	close();

	_tune=hvl_LoadTune((TEXT*)fileName.text(), _settings.samplerate, 4);

	if (!_tune)
		return false;

	hvl_InitSubsong(_tune, 0);

	addInfo(TITLE, (char*)_tune->ht_Name);
	FXString comment=FXString::null;
	for (unsigned int i=0; i<_tune->ht_InstrumentNr; i++)
		comment.append(FXStringFormat("%s\n", _tune->ht_Instruments[i].ins_Name));

	comment.trim();
	addInfo("Comment", comment.text());

	_left=0;
	_size=(_settings.samplerate*4)/50;
	_buffer=new char[_size];

	songLoaded(true);
	return true;
}

void hvl::close() {

	if (_tune) {
		hvl_FreeTune(_tune);
		delete[] _buffer;
		_buffer=0;
	}

	songLoaded(false);
}

const bool hvl::rewind() {

	if (_tune) {
		hvl_InitSubsong(_tune, _tune->ht_SongNum);
		return true;
	}
	return false;
}

const bool hvl::config(FXWindow* win) {

	hvlConfig cWin(win, _settings);
	if (cWin.execute(PLACEMENT_OWNER)) {
		_settings=cWin.settings();
		return true;
	}
	return false;
}

bool hvl::subtune(const FXuint i) {

	if (_tune) {
		hvl_InitSubsong(_tune, i);
		return true;
	}
	return false;
}

/*****************************************************************/

FXDEFMAP(hvlConfig) hvlConfigMap[] = {
	// Message_Type		ID				Message_Handler
	FXMAPFUNC(SEL_COMMAND,	hvlConfig::ID_SAMPLERATE,	hvlConfig::onCmdSamplerate),
};

FXIMPLEMENT(hvlConfig, configWindow, hvlConfigMap, ARRAYNUMBER(hvlConfigMap))

hvlConfig::hvlConfig(FXWindow* win, hvlConfig_t settings) :
	configWindow(win, hvl::name, CREDITS, LINK),
	_settings(settings) {

	FXint val;

	new FXLabel(matrix(), tr("Samplerate"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
	FXListBox *freqBox=new FXListBox(matrix(), this, hvlConfig::ID_SAMPLERATE, LAYOUT_FILL|FRAME_NORMAL);
	freqBox->fillItems("11025\n22050\n44100\n48000");
	freqBox->setNumVisible(4);
	switch (_settings.samplerate) {
	case 11025:
		val=0;
		break;
	case 22050:
		val=1;
		break;
	default:
	case 44100:
		val=2;
		break;
	case 48000:
		val=3;
		break;
	}
	freqBox->setCurrentItem(val);
}

long hvlConfig::onCmdSamplerate(FXObject*, FXSelector, void* ptr) {

	const FXint val=(FXint)(FXival)ptr;
	switch (val) {
	case 0:
		_settings.samplerate=11025;
		break;
	case 1:
		_settings.samplerate=22050;
		break;
	case 2:
		_settings.samplerate=44100;
		break;
	case 3:
		_settings.samplerate=48000;
		break;
	}
	return 1;
}


#endif
