/*
 *  Copyright (C) 2006-2009 Leandro Nini
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef MOD_H
#define MOD_H

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#ifdef HAVE_LIBMODPLUG
#include <libmodplug/stdafx.h>
#include <libmodplug/sndfile.h>

#include "inputPlugin.h"
#include <fx.h>
#if defined(HAVE_MMAP) || defined(_WIN32)
#  include <FXMemMap.h>
#endif

/*****************************************************************/

typedef struct {
	FXbool	oversampling;
	FXbool	noiseRed;
	FXint	channels;
	FXint	bits;
	FXint	samplerate;
	FXint	resamplingMode;
	FXint	reverbDepth;
	FXint	reverbDelay;
	FXint	bassAmount;
	FXint	bassRange;
	FXint	surroundDepth;
	FXint	surroundDelay;
} modConfig_t;

/*****************************************************************/

class mod : public inputPlugin {

private:
	modConfig_t	_settings;
	CSoundFile	*_module;

	unsigned int	_sampleSize;

	const bool	_mmx;
	bool		_unmo3;

	FXString	_ext;

private:
	/// Get temp file name
	const FXString tempFile(const FXString& fileName) {
		return FXSystem::getTempDirectory()+PATHSEP+FXPath::title(fileName)+".mod";
	}

	/// Delete temp file
	void delTempFile(const bool del, const FXString& tmpFile) {
		if (del) {
			FXTRACE((10, "Deleting temp file: %s\n", tmpFile.text()));
			FXFile::remove(tmpFile);
		}
	}

	void setSampleSize() {
		_sampleSize=_settings.channels*(_settings.bits>>3);
	}

public:
	mod(FXApp *a);
	~mod();

	static const char name[];

	/// Factory method
	static input* factory(FXApp* a) { return new mod(a); }

	/// Check if we support ext
	static bool supports(const char* fileName);

	/// Get supported extension
	FXString ext() const { return _ext; }

	/// Open file
	bool open(const FXString& fileName);

	/// Close file
	void close();

	/// Rewind to start
	const bool rewind();

	/// Open config dialog
	const bool config(FXWindow* win);

	/// Get samplerate
	const FXint samplerate() const { return _settings.samplerate; }

	/// Get channels
	const FXint channels() const { return _settings.channels; }

	/// Get precision
	const FXint precision() const { return _settings.bits>>3; }

	/// Callback function
	int fillBuffer(char* buffer, const unsigned int bufferSize, const unsigned int seconds);
};

/*****************************************************************/

#include "configWindow.h"

class modConfig : public configWindow {
	FXDECLARE(modConfig)

private:
	modConfig_t	_settings;

	FXDataTarget	_oversampling;
	FXDataTarget	_noiseRed;
	FXDataTarget	_revDepth;
	FXDataTarget	_revDelay;
	FXDataTarget	_bassAmount;
	FXDataTarget	_bassRange;
	FXDataTarget	_surrDepth;
	FXDataTarget	_surrDelay;

private:
	modConfig() {};
	modConfig(const modConfig&);
	modConfig& operator=(const modConfig&);

public:
	modConfig(FXWindow* win, modConfig_t settings);
	virtual ~modConfig() {};

	/// Get settings
	modConfig_t settings() const { return _settings; }

public:
	enum{
		ID_FREQUENCY=FXDialogBox::ID_LAST,
		ID_CHANNELS,
		ID_BITS,
		ID_RESAMPLING,
		ID_REVERB,
		ID_MEGABASS,
		ID_SURROUND,
		ID_LAST,
	};

public:
	long onCmdFrequency(FXObject*, FXSelector, void*);
	long onCmdChannels(FXObject*, FXSelector, void*);
	long onCmdBits(FXObject*, FXSelector, void*);
	long onCmdResampling(FXObject*, FXSelector, void*);
};

#endif

#endif
