/*
 *  Copyright (C) 2006-2009 Leandro Nini
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "ogg.h"

#ifdef HAVE_VORBIS

#include "utils.h"

#include <FXUTF8Codec.h>

ov_callbacks ogg::vorbis_callbacks = {
    ogg::read_func,
    ogg::seek_func,
    ogg::close_func,
    ogg::tell_func
};

#define EXT	"*.ogg"

#define CREDITS		"OggVorbis\nCopyright \302\251 Xiph.org Foundation"
#define LINK		"http://www.xiph.org/"

const char ogg::name[]="Ogg-Vorbis";

/*****************************************************************/

int ogg::fillBuffer(char* buffer, const unsigned int bufferSize, const unsigned int seconds) {

	unsigned int n=0;
	long read;
	do {
		int current_section;
		read=ov_read(_vf, buffer+n, bufferSize-n, 0,
			_settings.precision, (_settings.precision!=1), &current_section);
		n+=read;
	} while (read && n<bufferSize);

	return n;
}

/*****************************************************************/

bool ogg::supports(const char* fileName) {

	return FXPath::match(FXString(EXT), FXPath::name(fileName), FILEMATCH_FILE_NAME|FILEMATCH_NOESCAPE|FILEMATCH_CASEFOLD);
}

inline FXString ogg::ext() const { return FXString(EXT); }

ogg::ogg(FXApp* a) :
	inputPlugin(a, name),
	_vf(0),
	_vi(0) {

	_settings.precision=(load("Bits", 16)>>3);
}

ogg::~ogg() {

	close();

	save("Bits", (_settings.precision<<3));
}

bool ogg::open(const FXString& fileName) {

	close();

	if (!_file.open(fileName))
		return false;

	_vf=new OggVorbis_File;
	if (ov_open_callbacks(&_file, _vf, NULL, 0, vorbis_callbacks)<0) {
		delPtr(_vf);
		_file.close();
		return false;
	}

	_vi=ov_info(_vf, -1);

	time((FXint)ov_time_total(_vf, -1));

	FXString title=FXString::null;
	FXString artist=FXString::null;
	FXString year=FXString::null;
	FXString album=FXString::null;
	FXString genre=FXString::null;
	FXString comment=FXString::null;

	char **ptr=ov_comment(_vf,-1)->user_comments;
	while (*ptr) {
		FXTRACE((10, "%s\n", *ptr));
		if (!getComment(*ptr, &title, "title"))
		if (!getComment(*ptr, &artist, "artist"))
		if (!getComment(*ptr, &year, "date"))
		if (!getComment(*ptr, &album, "album"))
		if (!getComment(*ptr, &genre, "genre"))
		if (!getComment(*ptr, &comment, "comment")) {
			if (!comparecase(*ptr, "tracknumber", 11))
				addInfo(TRACK, FXString(*ptr).erase(0, 12).text());
		}
		++ptr;
	}

	addInfo(TITLE, title.text());
	addInfo(ARTIST, artist.text());
	addInfo(ALBUM, album.text());
	addInfo(GENRE, genre.text());
	addInfo(YEAR, year.text());
	addInfo(COMMENT, comment.text());

	songLoaded(true);
	return true;
}

bool ogg::getComment(const char* orig, FXString* dest, const char* type) {

	const int n=strlen(type);
	if (comparecase(orig, type, n))
		return false;

	if (!dest->empty())
		dest->append(", ");

	const FXUTF8Codec codec;
	dest->append(codec.mb2utf(orig+n+1));
	return true;
}

void ogg::close() {

	if (_vf) {
		ov_clear(_vf);
		delPtr(_vf);
		_file.close();
	}

	_vi=0;

	songLoaded(false);
}

const bool ogg::rewind() {

	if (!_vf)
		return false;

	if (ov_raw_seek(_vf, 0))
		return false;

	return true;
}

const bool ogg::config(FXWindow* win) {

	oggConfig cWin(win, _settings);
	if (cWin.execute(PLACEMENT_OWNER)) {
		_settings=cWin.settings();
		return true;
	}
	return false;
}

size_t ogg::read_func(void *ptr, size_t size, size_t nmemb, void *datasource) {

	return ((FXFile*)datasource)->readBlock((char*)ptr, size*nmemb);
}

int ogg::seek_func(void *datasource, ogg_int64_t offset, int whence) {

	FXuint from;
	switch (whence) {
	case SEEK_SET:
		from=FXIO::Begin;
		break;
	case SEEK_CUR:
		from=FXIO::Current;
		break;
	case SEEK_END:
		from=FXIO::End;
		break;
	default:
		return -1;
	}
	return ((FXFile*)datasource)->position(offset, from);
}

int ogg::close_func(void *datasource) {

	((FXFile*)datasource)->close();
	return 0;
}

long ogg::tell_func(void *datasource) {

	return ((FXFile*)datasource)->position();
}

/*****************************************************************/

FXDEFMAP(oggConfig) oggConfigMap[] = {
	// Message_Type		ID				Message_Handler
	FXMAPFUNC(SEL_COMMAND,	oggConfig::ID_BITS,		oggConfig::onCmdBits),
};

FXIMPLEMENT(oggConfig, configWindow, oggConfigMap, ARRAYNUMBER(oggConfigMap))

oggConfig::oggConfig(FXWindow* win, oggConfig_t settings) :
	configWindow(win, ogg::name, CREDITS, LINK),
	_settings(settings) {

	FXint val;

	new FXLabel(matrix(), tr("Bits"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
	FXListBox* _bitBox=new FXListBox(matrix(), this, oggConfig::ID_BITS, LAYOUT_FILL|FRAME_NORMAL);
	_bitBox->fillItems("8\n16");
	_bitBox->setNumVisible(2);
	switch (_settings.precision) {
	case 1:
		val=0;
		break;
	default:
	case 2:
		val=1;
		break;
	}
	_bitBox->setCurrentItem(val);
}

long oggConfig::onCmdBits(FXObject*, FXSelector, void* ptr) {

	const FXint val=(FXint)(FXival)ptr;
	switch (val) {
	case 0:
		_settings.precision=1;
		break;
	case 1:
		_settings.precision=2;
		break;
	}
	return 1;
}

#endif
