/*
 *  Copyright (C) 2006-2009 Leandro Nini
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "alsa.h"

#ifdef HAVE_ALSA

#include <stdio.h>
#include <stdlib.h>
#include <iostream>

const char alsa::name[]="ALSA";

/*****************************************************************/

alsa::alsa(FXApp* app) :
	outputPlugin(app, name),
	_pcmHandle(0),
	_mixer(0),
	_buffer(0) {

	// Check devices
	int card=-1;
	snd_card_next(&card);
	char* devName;
	while (card>=0) {
		snd_card_get_name(card, &devName);
		addDevice(devName);
		snd_card_next(&card);
	}
}

int alsa::open(const int card, int &sampleRate, const int channels, const int prec) {

	const FXString devName=FXStringFormat("hw:%d", card);
	FXTRACE((10,"Opening device: %s\n", devName.text()));
	const int err=snd_pcm_open(&_pcmHandle, devName.text(), SND_PCM_STREAM_PLAYBACK, 0);
	if (err<0) {
		fxmessage("Error: %s\n", snd_strerror(err));
		return 0;
	}

	const int bufSize=setParams(sampleRate, channels, prec);
	if (!bufSize) {
		snd_pcm_close(_pcmHandle);
		_pcmHandle=0;
		return 0;
	}

	initMixer(devName.text());

	return bufSize;
}

void alsa::close() {

	//snd_pcm_drain(_pcmHandle);
	snd_pcm_drop(_pcmHandle);
	snd_pcm_close(_pcmHandle);
	_pcmHandle=0;

	if (_mixer) {
		snd_mixer_close(_mixer);
		_mixer=0;
	}

	delete [] _buffer;
}

const int alsa::setParams(int &sampleRate, const int channels, const int prec) {

	int err;

	snd_pcm_hw_params_t *hwParams;
	snd_pcm_hw_params_alloca(&hwParams);

	err=snd_pcm_hw_params_any(_pcmHandle, hwParams);
	if (err<0) {
		fxmessage("Error: %s\n", snd_strerror(err));
		return 0;
	}

	err=snd_pcm_hw_params_set_access(_pcmHandle, hwParams, SND_PCM_ACCESS_RW_INTERLEAVED);
	if (err<0) {
		fxmessage("Error: %s\n", snd_strerror(err));
		return 0;
	}

	// Set sample format
	snd_pcm_format_t format;
	switch (prec) {
	case 1:
		format=SND_PCM_FORMAT_U8;
		break;
	case 2:
		format=SND_PCM_FORMAT_S16_LE;
		break;
	case 3:
		format=SND_PCM_FORMAT_S24_LE;
		break;
	case 4:
		format=SND_PCM_FORMAT_S32_LE;
		break;
	default:
		fxmessage("Unsupported format\n");
		return 0;
	}

	err=snd_pcm_hw_params_set_format(_pcmHandle, hwParams, format);
	if (err<0) {
		fxmessage("Error: %s\n", snd_strerror(err));
		return 0;
	}

	// Set number of channels
	err=snd_pcm_hw_params_set_channels(_pcmHandle, hwParams, channels);
	if (err<0) {
		fxmessage("Error: %s\n", snd_strerror(err));
		return 0;
	}

	// Set samplerate
	err=snd_pcm_hw_params_set_rate_near(_pcmHandle, hwParams, (unsigned int*)&sampleRate, 0);
	if (err<0) {
		fxmessage("Error: %s\n", snd_strerror(err));
		return 0;
	}

	// Set buffer size to 1/4 of second
	snd_pcm_uframes_t bufSize=sampleRate>>2;
	err=snd_pcm_hw_params_set_buffer_size_near(_pcmHandle, hwParams, &bufSize);
	if (err<0) {
		fxmessage("Error: %s\n", snd_strerror(err));
		return 0;
	}

	err=snd_pcm_hw_params(_pcmHandle, hwParams);
	if (err<0) {
		fxmessage("Error: %s\n", snd_strerror(err));
		return 0;
	}

	int bufferSize=snd_pcm_frames_to_bytes(_pcmHandle, bufSize);
	FXTRACE((10, "Buffer size=%d bytes\n", bufferSize));

	err=snd_pcm_prepare(_pcmHandle);
	if (err<0) {
		fxmessage("Error: %s\n", snd_strerror(err));
		return 0;
	}

	_buffer=new char[bufferSize];

	return bufferSize;
}

bool alsa::write(char* buffer, int bufferSize) {

	bufferSize=snd_pcm_bytes_to_frames(_pcmHandle, bufferSize);
loop:
	int n=snd_pcm_writei(_pcmHandle, buffer, bufferSize);

	if (n<0) {
		n=snd_pcm_recover(_pcmHandle, n, 0);
		if (n<0)
			return false;
		goto loop;
	}

	bufferSize-=n;
	if (!bufferSize)
		return true;

	buffer+=snd_pcm_frames_to_bytes(_pcmHandle, n);
	goto loop;
}

void alsa::volume(int vol) {

	if (!_mixer)
		return;

	if (snd_mixer_selem_set_playback_volume_all(_mixerHandle, vol)<0)
		FXTRACE((10, "Error setting volume\n"));
}

bool alsa::initMixer(const char* dev) {

	int err;
	long min, max;

	err=snd_mixer_open(&_mixer, 0);
	if (err<0)
		goto errorMsg;

	err=snd_mixer_attach(_mixer, dev);
	if (err<0)
		goto errorMsg;

	err=snd_mixer_selem_register(_mixer, 0, 0);
	if (err<0)
		goto errorMsg;

	err=snd_mixer_load(_mixer);
	if (err<0)
		goto errorMsg;

	snd_mixer_selem_id_t *selem_id;
	err=snd_mixer_selem_id_malloc(&selem_id);
	if (err<0)
		goto errorMsg;

	snd_mixer_selem_id_set_index(selem_id, 0);
	snd_mixer_selem_id_set_name(selem_id, "PCM");

	_mixerHandle=snd_mixer_find_selem(_mixer, selem_id);

	if (!snd_mixer_selem_has_playback_volume(_mixerHandle)) {
		snd_mixer_selem_id_set_name(selem_id, "Wave");
		_mixerHandle=snd_mixer_find_selem(_mixer, selem_id);
		if (!snd_mixer_selem_has_playback_volume(_mixerHandle))
			_mixerHandle=0;
	}

	snd_mixer_selem_id_free(selem_id);

	if (!_mixerHandle)
		goto error;

	snd_mixer_selem_get_playback_volume_range(_mixerHandle, &min, &max);

	if (!max)
		goto error;

	snd_mixer_selem_set_playback_volume_range(_mixerHandle, 0, 100);

	snd_mixer_selem_get_playback_volume(_mixerHandle, SND_MIXER_SCHN_FRONT_LEFT, &min);
	snd_mixer_selem_get_playback_volume(_mixerHandle, SND_MIXER_SCHN_FRONT_RIGHT, &max);

	return true;

errorMsg:
	fxmessage("Error: %s\n", snd_strerror(err));
error:
	if (_mixer) {
		snd_mixer_close(_mixer);
		_mixer=0;
	}
	return false;
}

#endif
