/*
 *  Copyright (C) 2006-2009 Leandro Nini
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "utils.h"

#ifdef _WIN32
#include <windows.h>

// Based on code and info by Shane Hird (at www.msnfanatic.com) and Marijn Kentie (kentie.net)

void nowPlaying(const bool show, const bool wlm, const FXString& title, const FXString& artist, const FXString& album) {

	FXString message=FXStringFormat("\\0Music\\0%d\\0%s%s\\0%s\\0%s\\0", show?1:0,
		wlm?"{0} - {1}\\0":"", title.text(), artist.text(), album.text());
	FXTRACE((10, "nowPlaying %s\n", message.text()));

	const FXint n=ncslen(message.text(), message.length());
	FXnchar* buffer=new FXnchar[n+1];
	utf2ncs(buffer, message.text(), message.length()+1);

	COPYDATASTRUCT msndata;

	msndata.dwData = 0x547;
	msndata.lpData = (void*)buffer;
	msndata.cbData = n*2+2;

	HWND msnui=NULL;
#ifdef UNICODE
	while (msnui=FindWindowEx(NULL, msnui, L"MsnMsgrUIManager", NULL))
#else
	while (msnui=FindWindowEx(NULL, msnui, "MsnMsgrUIManager", NULL))
#endif
		SendMessage(msnui, WM_COPYDATA, (WPARAM)NULL, (LPARAM)&msndata);

	delete [] buffer;
}
#endif

void runCmd(const FXString& cmd) {
#ifdef _WIN32
	PROCESS_INFORMATION pi={0};
	STARTUPINFO si={sizeof(si)};
	si.dwFlags=STARTF_USESHOWWINDOW;
	si.wShowWindow=SW_HIDE;

#ifdef UNICODE
	const FXint n=ncslen(cmd.text(), cmd.length());
	FXnchar* command=new FXnchar[n+1];
	utf2ncs(command, cmd.text(), cmd.length()+1);

	if (CreateProcess(0, command, NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi)) {
#else
	if (CreateProcess(0, const_cast<char*>(cmd.text()), NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi)) {
#endif //UNICODE
		WaitForSingleObject(pi.hProcess, INFINITE);
		CloseHandle(pi.hProcess);
		CloseHandle(pi.hThread);
	}
#ifdef UNICODE
	delete [] command;
#endif //UNICODE
#else
	system(cmd.text());
#endif //_WIN32
}

#define MAX_CHARS	20

FXString shrink(const FXString& string) {

	return (string.count()>MAX_CHARS)?
		FXStringFormat("...%s", string.right(MAX_CHARS).text()):
		string;
}

#define MAXBUFFER	256
#define EXT		"*.pls,*.m3u"

FXString plExt() { return FXString(EXT); }

FXint parsePls(FXString* &filelist, const FXString &path) {

	FXSettings pls;
	pls.parseFile(path, FALSE);

	FXint items=pls.readIntEntry("playlist", "NumberOfEntries");
	filelist=new FXString[items];

	FXint count=0;
	while (items--) {
		FXString file=pls.readStringEntry("playlist", FXStringFormat("File%d", items+1).text());
		if (FXStat::exists(file))
			filelist[count++]=file;
	}

	return count;
}

FXint parseM3u(FXString* &filelist, const FXString &path) {

	FXString *newlist;
	FXint size=0;
	FXint count=0;

	FXchar buffer[MAXBUFFER];
	FILE* file=fopen(path.text(), "r");

	while (fgets(buffer, MAXBUFFER, file)) {
		if (buffer[0]!='#') {
			int i=1;
			while (Ascii::isPrint(buffer[i])) i++;
			buffer[i]='\0';

// code taken from FXDir.cpp which is part of the FOX toolkit (c) Jeroen van der Zijp
      // Grow list
      if(count+1>=size){
        size=size?(size<<1):256;
        newlist=new FXString [size];
        for(int i=0; i<count; i++){
          newlist[i].adopt(filelist[i]);
          }
        delete [] filelist;
        filelist=newlist;
        }

			filelist[count++]=FXString(buffer);
		}
	}
	fclose(file);

	return count;
}

FXint plLoad(FXString* &filelist, const FXString &path) {

	filelist=0;

	const FXString ext=FXPath::extension(path);

	if (!comparecase(ext, "pls"))
		return parsePls(filelist, path);

	if (!comparecase(ext, "m3u"))
		return parseM3u(filelist, path);

	return 0;
}

void writeLine(FILE* file, FXString line) {

	line.append(ENDLINE);
	fputs(line.text(), file);
}

bool saveM3u(FXString* &filelist, FXint num, const FXString &path) {

	FILE* file=fopen(path.text(), "w");
	if (!file)
		return false;

	for (int i=0; i<num; i++) {
		writeLine(file, filelist[i]);
	}

	fclose(file);
	return true;
}

bool savePls(FXString* &filelist, FXint num, const FXString &path) {

	FILE* file=fopen(path.text(), "w");
	if (!file)
		return false;

	writeLine(file, "[playlist]");

	FXString tmp=FXStringFormat("NumberOfEntries=%d", num);
	writeLine(file, tmp);

	for (int i=0; i<num; i++) {
		tmp=FXStringFormat("File%d=%s", i+1, filelist[i].text());
		writeLine(file, tmp);
	}

	fclose(file);
	return true;
}

bool plSave(FXString* &filelist, FXint num, const FXString &path) {

	const FXString ext=FXPath::extension(path);

	if (!comparecase(ext, "pls"))
		return savePls(filelist, num, path);

	if (!comparecase(ext, "m3u"))
		return saveM3u(filelist, num, path);

	return false;
}
